
import numpy as np
from scipy.stats import norm


class PreferenceExpectedImprovement:
    def __init__(self, nr_dims, initial_variance, update_variance, nr_samples=100,
                 kappa=0.0, lower_bound=None, upper_bound=None, seed=None, fixed_dims=None):
        self.nr_dims = nr_dims

        self.initial_variance = initial_variance
        self.update_variance = update_variance

        self.nr_samples = nr_samples
        self.kappa = kappa

        if lower_bound is None:
            self.lower_bound = [-1.] * self.nr_dims
        else:
            self.lower_bound = lower_bound

        if upper_bound is None:
            self.upper_bound = [1.] * self.nr_dims
        else:
            self.upper_bound = upper_bound

        self.seed = seed

        # initial proposal distribution
        self.proposal_mean = np.zeros((nr_dims, 1))
        self.proposal_cov = np.diag(np.ones((nr_dims,)) * self.initial_variance)

        # fixed dimension for robot experiment
        self.fixed_dims = fixed_dims

    def rejection_sampling(self, seed=None):
        rng = np.random.default_rng(seed)

        samples = np.empty((0, self.nr_dims))
        while samples.shape[0] < self.nr_samples:
            # sample from the multi variate gaussian distribution
            sample = np.zeros((1, self.nr_dims))
            for i in range(self.nr_dims):
                if i in self.fixed_dims:
                    sample[0, i] = self.fixed_dims[i]
                else:
                    check = False
                    while not check:
                        sample[0, i] = rng.normal(self.proposal_mean[i], self.proposal_cov[i, i])
                        if self.lower_bound[i] <= sample[0, i] <= self.upper_bound[i]:
                            check = True

            samples = np.append(samples, sample, axis=0)

        return samples

    def __call__(self, gauss_process, x_observed, seed=None):
        # if seed is set for whole experiment
        if self.seed is not None:
            seed = self.seed

        # get the best so far observed y
        mu = gauss_process.predict(x_observed)
        y_best = max(mu)

        # sample from surrogate
        x_test = self.rejection_sampling(seed)
        mu, sigma = gauss_process.predict(x_test, return_std=True)

        # expected improvement
        z = (mu - y_best - self.kappa) / sigma
        ei = (mu - y_best - self.kappa) * norm.cdf(z) + sigma * norm.pdf(z)

        # get the best result and return it
        idx = np.argmax(ei)
        return x_test[idx, :]

    def update_proposal_model(self, preference_mean, preference_bool):
        cov_diag = np.ones((self.nr_dims,)) * self.initial_variance
        cov_diag[preference_bool] = self.update_variance

        preference_cov = np.diag(cov_diag)

        preference_mean = preference_mean.reshape(-1, 1)

        posterior_mean = np.linalg.inv(np.linalg.inv(self.proposal_cov) + np.linalg.inv(preference_cov))\
            .dot(np.linalg.inv(self.proposal_cov).dot(self.proposal_mean)
                 + np.linalg.inv(preference_cov).dot(preference_mean))

        posterior_cov = np.linalg.inv(np.linalg.inv(self.proposal_cov) + np.linalg.inv(preference_cov))

        self.proposal_mean = posterior_mean
        self.proposal_cov = posterior_cov
